/**************************************************************************/
/*                                                                        */
/*  This file is part of Frama-Clang                                      */
/*                                                                        */
/*  Copyright (C) 2012-2018                                               */
/*    CEA (Commissariat à l'énergie atomique et aux énergies              */
/*         alternatives)                                                  */
/*                                                                        */
/*  you can redistribute it and/or modify it under the terms of the GNU   */
/*  Lesser General Public License as published by the Free Software       */
/*  Foundation, version 2.1.                                              */
/*                                                                        */
/*  It is distributed in the hope that it will be useful,                 */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*  GNU Lesser General Public License for more details.                   */
/*                                                                        */
/*  See the GNU Lesser General Public License version 2.1                 */
/*  for more details (enclosed in the file LICENSE).                      */
/*                                                                        */
/**************************************************************************/

//
// Description:
//   Definition of the ACSL++ lexer.
//

#ifndef ACSL_LexerH

#include <list>
#include "DescentParse.h"
#include "ACSLToken.h"

extern "C" {

#include "intermediate_format.h"

}

/** @file */

namespace Acsl {

/*! @class Lexer
 *  @brief Builds a token Token from a string buffer.
 *
 *  The main method of the class Lexer is readToken. It reads characters from
 *    the input string buffer and then it returns the token that has been read.
 *    For that it uses a state machine where the states are method pointers,
 *    that is a pointer to one of the following methods: readMain,
 *    readEndComment, readIdentifierToken, readNumberToken, readProtectedToken.
 *    \n \n
 *  The token is completly read when readToken returns RRHasToken. Then it
 *    is available through the method queryToken(). If queryToken() reveals to
 *    be a first kind enum token, then a simple static_cast conversion
 *    gives an access to what it is. If queryToken() reveals to be a second kind
 *    elaborate token the method getContentToken gives a read access to its
 *    content. \n \n
 *  The Lexer is embedded in the Parser class as well as in the
 *    Parser::Arguments class. For the Parser class it helps to implement
 *    the Parser::parse method. For the Parser::Arguments, it helps to implement
 *    an inlined access to Parser::Arguments::setToNextToken enabling a more
 *    efficient Parser::TStateStack::parse implementation.
 */
class Lexer : public Parser::Base {
public:
  /*! @class Error
   *  @brief Defines a lexing error to be reported into a Acsl::ErrorMessage.
   */
  class Error {
  public:
    std::string filePos;
    unsigned linePos;
    unsigned columnPos;
    std::string message;

    Error(location position, const std::string& msg);
    Error(const std::string& file, unsigned line, unsigned column,
        const std::string& msg)
      : filePos(file), linePos(line), columnPos(column), message(msg) {}
  };
  typedef DLexer::AbstractToken AbstractToken;

private:
  typedef ReadResult (Lexer::*ReadPointerMethod)(const std::string& buffer,
      size_t& position, location loc, bool doesForce);
  typedef DLexer::Token Token;

  typedef Parser::TTextBuffer<char> TextBuffer;
  typedef std::list<std::pair<std::string,AbstractToken> > Utf8SymbolSet;
  TextBuffer _currentToken;
    //!< content of _token that the lexer is currently reading.
  ReadPointerMethod _mainState, _secondaryState; 
    //!< State of the state machine during the lexing.
  char _context;   //!< context for comments and for constant reading.
  char _extension; //!< additional information for reading a literal extension.
  Token _token;    //!< Token that the lexer is currently reading.
  bool _hasNewlineToken; //!< if the previous token was a comment
  enum { Unknown, Dec, Oct, Hex, Bin } _digits;
    //!< kind of digits allowed when parsing numeric constant
  enum { LeadZero, Digits, FracPart, Exponent, ExpNext, ISuf, FSuf } _litState;
    //!< current state of numeric literal parsing
  DLexer::CharacterLiteralToken::Type _charLitKind;
  std::list<Error> _errorList; //!< Error list encountered by the lexer.

  //! utf8-encoded symbols that are parsed in ACSL specifications
  Utf8SymbolSet _AcceptedUtf8Symbols;

  //! converts an unicode character into its utf8 representation
  static std::string utf8_convert(unsigned int v);

  /*! whether the given character could be the first character of an
      accepted utf8 symbol.
   */
  bool isUtf8SymbolStart(char ch);

  // put current char in current token and advance to next char.
  void advanceChar(const std::string& buffer, size_t& position, location loc) {
    _currentToken << buffer[position];
    position++;
    loc->charnum2++;
  }

  // returns true if the character is a valid digit for the currently
  // selected base. Assumes that _digits is not UNKNOWN.
  bool isDigit(char ch) {
    assert(_digits);
    if (ch == '0' || ch == '1') return true;
    if (_digits == Bin) return false;
    if ( '0' <= ch && ch <= '7') return true;
    if (_digits == Oct) return false;
    if (ch == '8' || ch == '9') return true;
    if (_digits == Dec) return false;
    if ('a' <= ch && ch <= 'f') return true;
    return ('A' <= ch && ch <= 'F');
  }

  void setType(DLexer::IntegerLiteralToken *intToken) const {
    if (_digits == Hex)
      intToken->setType(DLexer::IntegerLiteralToken::THexaDecimal);
    if (_digits == Bin)
      intToken->setType(DLexer::IntegerLiteralToken::TBit);
    if (_digits == Oct)
      intToken->setType(DLexer::IntegerLiteralToken::TOctal);
  }

protected:
  //! read an AbstractToken from the beginning and stores it in _token.
  ReadResult readMain(const std::string& buffer, size_t& position, location loc,
      bool doesForce);
  //! read the end of a CommentToken, depending of the kind of comment
  //! (a line one or a delimited one).
  ReadResult readEndComment(const std::string& buffer, size_t& position,
      location loc, bool doesForce);
  //! read the end of an IdentifierToken from buffer, position.
  ReadResult readIdentifierToken(const std::string& buffer, size_t& position,
      location loc, bool doesForce);
  //! read the end of a LiteralToken from buffer, position.
  ReadResult readNumberToken(const std::string& buffer, size_t& position,
      location loc, bool doesForce);

  //! read a character literal
  ReadResult readCharLiteral(
    const std::string& buffer, size_t& position, location loc, bool doesForce);

  //! read the end of an IdentifierToken starting with '\\'.
  ReadResult readProtectedToken(const std::string& buffer, size_t& position,
      location loc, bool doesForce);
  ReadResult readUtf8Symbol(const std::string& buffer, size_t& position,
                            location loc, bool doesForce);
public:
  Lexer()
    : _mainState(&Lexer::readMain), _secondaryState(NULL), _context('\0'),
      _extension(0), _hasNewlineToken(false),
      _charLitKind(DLexer::CharacterLiteralToken::TUndefined),
      _AcceptedUtf8Symbols()
      {}

  Lexer(const Lexer& source)
    : _currentToken(source._currentToken), _mainState(source._mainState),
      _secondaryState(source._secondaryState), _context(source._context),
      _extension(source._extension), _token(source._token),
    _hasNewlineToken(source._hasNewlineToken),
    _charLitKind(source._charLitKind),
    _AcceptedUtf8Symbols()
    { }

  ReadResult readToken(const std::string& buffer, size_t& position, location
      loc, bool doesForce=false)
    { return _secondaryState
          ? (this->*_secondaryState)(buffer, position, loc, doesForce)
          : (this->*_mainState)(buffer, position, loc, doesForce);
    }
  void setToken(const Token& token)
    { assert(_token.getType() == AbstractToken::TUndefined);
      _token = token;
    }
  AbstractToken queryToken() { return _token.getFullToken(); }
  const AbstractToken& getContentToken() const { return _token.getContent(); }
  void assumeContentToken() { _token.assumeContent(); }
  bool hasErrors() const { return !_errorList.empty(); }
  std::list<Error>& errorList() { return _errorList; }
  Token& extractToken() { return _token; }
  bool doesNeedClear() const { return _token.getType(); }

  void clearToken()
    { _currentToken.clear();
      _mainState = &Lexer::readMain;
      _secondaryState = NULL;
      _context = '\0';
      _token = Token();
    }
};

} // end of namespace Acsl

#endif // ACSL_LexerH
